#include "Progress.h"
#include <QMetaType>
#include "globals.h"
#include <stdio.h>

// NOTE: {i}, {l:r}, etc. denotes total values
//       <i>, <l:r>, etc. denotes act number (if present)
//       [i], [l:r], etc. denotes thread number
//       (i), (l:r), etc. denotes item values
#undef DEBUG_PROGRESS
#undef DEBUG_PROGRESS_DETAIL

// With rate limiting, the minion's event queue is bombarded by so many
// events that it can't keep up.  It ends up with a huge backlog of events
// by the time the whole activity is complete and it is destroyed.  The
// result is that the progress bars indicate only a fraction of the total
// progress being complete when suddenly the activity is done.
//
// Checks in the most frequent event functions implement a rate limit that
// ensures that the minion can keep up.  No more than one update of those event
// functions is allowed every RATE_LIMIT_INTERVAL.  However, a special case will
// allow the last item value event through no matter what.  Once.
//
// For debugging, it can be useful to set this to 0 to see more progress info.
#define RATE_LIMIT_INTERVAL  10 // ms

// In a single-threaded implementation (not used by NUU), this is the interval
// with which the main event loop would be run.
#define GUI_UPDATE_INTERVAL  10 // ms

#ifdef DEBUG_PROGRESS
static int  indent = 0;
#endif

bool  Progress::Object::registeredTypes = false;

Progress::Object::Object(HProgressBar*  totalProgress,
                         HProgressBar*  itemProgress)
{
   if (!registeredTypes) {
      qRegisterMetaType<Steps>("Steps");
      qRegisterMetaType<ThreadsState>("ThreadsState");
      registeredTypes = true;
   }

#if defined(DEBUG_PROGRESS)
   printf("----------------------------------------"
          "----------------------------------------\n");
   printf("progress: create\n");
#endif

   minion = new Minion(totalProgress, itemProgress);

   minion->moveToThread(application->thread());

   singleThread = (minion->thread() == thread());

   connect(this,   SIGNAL(startTotalSig(QString, Steps)),
           minion, SLOT(startTotal(QString, Steps)));
   connect(this,   SIGNAL(changeTotalSig(Steps, Steps)),
           minion, SLOT(changeTotal(Steps, Steps)));
   connect(this,   SIGNAL(startItemSig(QString, Steps, Steps)),
           minion, SLOT(startItem(QString, Steps, Steps)));
   connect(this,   SIGNAL(changeItemActivitySig(QString)),
           minion, SLOT(changeItemActivity(QString)));
   connect(this,   SIGNAL(changeItemActsSig(Steps)),
           minion, SLOT(changeItemActs(Steps)));
   connect(this,   SIGNAL(changeItemActStepsSig(Steps, Steps, Steps)),
           minion, SLOT(changeItemActSteps(Steps, Steps, Steps)));
   connect(this,   SIGNAL(setItemValueSig(Steps, Steps)),
           minion, SLOT(setItemValue(Steps, Steps)));
   connect(this,   SIGNAL(endItemActsSig(Steps)),
           minion, SLOT(endItemActs(Steps)));
   connect(this,   SIGNAL(setTotalValueSig(Steps)),
           minion, SLOT(setTotalValue(Steps)));
   connect(this,   SIGNAL(updateThreadsStateSig(const ThreadsState&)),
           minion, SLOT(updateThreadsState(const ThreadsState&)));
   connect(this,   SIGNAL(resetSig()),
           minion, SLOT(reset()));

   reset();
}

Progress::Object::~Object()
{
   reset();

#if defined(DEBUG_PROGRESS)
   printf("progress: delete\n");
   printf("----------------------------------------"
          "----------------------------------------\n");
#endif

   minion->deleteLater();
}

void Progress::Object::startTotal(QString  activity,
                                  Steps    totalSteps)
{
   totalValue = 0;
   totalTail  = totalSteps;

#if defined(DEBUG_PROGRESS)
   printf("progress: startTotal: {0/%llu} %s\n",
          totalSteps, activity.toStdString().c_str());
   indent = 3;
#endif

   if (singleThread) {
      lastGuiUpdate.start();
   }

   emit startTotalSig(activity, totalTail);

   updateGui();
}

void Progress::Object::changeTotal(Steps  totalSteps)
{
   if (totalValue > totalTail) {
      totalValue = totalTail;
   }
   totalTail = totalSteps;

#if defined(DEBUG_PROGRESS)
   printf("progress: changeTotal: {%llu/%llu}\n",
          totalValue, totalSteps);
   indent = 3;
#endif

   emit changeTotalSig(totalValue, totalTail);

   updateGui();
}

void Progress::Object::startItem(QString  activity,
                                 Steps    totalSteps,
                                 Steps    itemSteps)
{
   totalValue = itemTotalTail;

   itemTotalStart = itemTotalTail;
   itemTotalTail  = itemTotalStart + totalSteps;

   itemValue = 0;
   itemTail  = itemSteps;

#if defined(DEBUG_PROGRESS)
   printf("progress:%*s startItem: {%llu:%llu (%llu)} (0:%llu) %s\n",
          indent, "",
          itemTotalStart, itemTotalTail, totalSteps, itemSteps,
          activity.toStdString().c_str());
   indent = 6;
#endif

   lastSetValue.start();

   emit startItemSig(activity, totalValue, itemTail);

   updateGui();
}

void Progress::Object::changeItemActivity (QString  activity)
{
#if defined(DEBUG_PROGRESS)
   printf("progress:%*s changeItemActivity: %s\n",
          indent, "", activity.toStdString().c_str());
#endif

   emit changeItemActivitySig(activity);

   updateGui();
}

void Progress::Object::setupItemActs (Steps  totalSteps,
                                      Steps  numActs)
{
   actsTotalSteps = totalSteps;
   actsTotalTail  = itemTotalTail + totalSteps;
   this->numActs  = numActs;
   prevAct        = 0;

#if defined(DEBUG_PROGRESS)
   printf("progress:    setupItemActs: {%llu:%llu (%llu)} <0:%llu>\n",
          actsTotalTail - actsTotalSteps, actsTotalTail, actsTotalSteps, 
          numActs);
   indent = 6;
#endif
}

void Progress::Object::changeItemActs (Steps  numActs)
{
   this->numActs = numActs;

   Steps  actsTotalStart    = actsTotalTail - actsTotalSteps;
   Steps  actsTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : prevAct * actsTotalSteps / numActs);
   itemTotalTail = actsTotalStart + actsTotalProgress;

#if defined(DEBUG_PROGRESS)
   printf("progress:    changeItemActs: <%llu> {%llu->%llu/%llu}\n", 
          numActs, totalValue, itemTotalTail, totalTail);
#endif

   totalValue = itemTotalTail;

   emit changeItemActsSig(totalValue);

   updateGui();
}

void Progress::Object::startItemAct (QString  activity,
                                     Steps    itemSteps,
                                     Steps    act)
{
   if (act > numActs) act = numActs;

#if defined(DEBUG_PROGRESS)
   printf("progress:       startItemAct: <%llu:%llu (%llu)> %s\n",
          prevAct, act, act - prevAct,
          activity.toStdString().c_str());
   indent = 6; // startItem called below should be at same indent
#endif

   Steps  prevTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : prevAct * actsTotalSteps / numActs);
   Steps  thisTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : act * actsTotalSteps / numActs);
   startItem(activity, thisTotalProgress - prevTotalProgress, itemSteps);
#if defined(DEBUG_PROGRESS)
   indent = 9;
#endif
   prevAct = act;
}

void Progress::Object::changeItemActSteps (Steps  value,
                                           Steps  itemSteps)
{
   // This includes the functionality of setItemValue, because it's so common
   // to update the number of steps only when also updating the value.

   if (value > itemSteps) {
      value = itemSteps;
   }
   
   itemValue = value;
   itemTail  = itemSteps;

   Steps  itemLength      = itemTail;
   Steps  itemTotalLength = itemTotalTail - itemTotalStart;

   totalValue = (itemLength == 0
                 ? itemTotalStart + itemTotalLength
                 : itemTotalStart + value * itemTotalLength / itemLength);

   lastSetValue.start();

#if defined(DEBUG_PROGRESS)
   printf("progress:%*s changeItemActSteps: {%llu/%llu}, (%llu/%llu)\n",
          indent, "",
          totalValue,
          totalTail,
          value,
          itemTail);
#endif

   emit changeItemActStepsSig(totalValue, value, itemTail);

   updateGui();
}

void Progress::Object::endItemActs()
{
   // Sometimes the trailing act that increment the number of acts to numActs
   // will not be performed.  In case that happens, ensure the total progress
   // is updated.  This is harmless if the last act was performed.
   itemTotalStart = actsTotalTail;
   itemTotalTail  = actsTotalTail;

   totalValue = itemTotalStart;

#if defined(DEBUG_PROGRESS)
   printf("progress:    endItemActs: {%llu/%llu}\n", actsTotalTail, totalTail);
   indent = 3;
#endif

   emit endItemActsSig(totalValue);

   updateGui();
}

void Progress::Object::setupItemActsThreads (QString  activity,
                                             Steps    firstAct,
                                             Steps    tailAct,
                                             Threads  numThreads,
                                             Steps    itemSteps)
{
#if defined(DEBUG_PROGRESS)
   printf("progress:       setupItemActsThreads: <%llu:%llu> [0:%d] %s\n",
          firstAct, tailAct, numThreads,
          activity.toStdString().c_str());
   indent = 6; // startItem called below should be at same indent
#endif

   threadsFirstAct  = firstAct;
   threadsNumActs   = numActs;

   threads = ThreadsState(numThreads);
   emit updateThreadsStateSig(threads);

   Steps  prevTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : prevAct * actsTotalSteps / numActs);
   Steps  thisTotalProgress = (numActs == 0
                               ? actsTotalSteps
                               : (tailAct-1) * actsTotalSteps / numActs);
   startItem(activity, thisTotalProgress - prevTotalProgress, itemSteps);
#if defined(DEBUG_PROGRESS)
   indent = 9;
#endif
   prevAct = tailAct-1;
}

void Progress::Object::startItemActThread (QString  activity,
                                           Threads  thread,
                                           Steps    itemSteps)
{
#if defined(DEBUG_PROGRESS)
   printf("progress:          startItemActThread: thread = [%d] (0:%llu) %s\n",
          thread, itemSteps, activity.toStdString().c_str());
   indent = 12;
#endif

   threads[thread].activity = activity;
   threads[thread].value    = 0;
   threads[thread].tail     = itemSteps;
   threads[thread].state    = Progress::Running;

   if (threads.count() == 1) {
      // Degenerate parallelism with only 1 thread.  Change the name to reflect
      // the single activity.
      changeItemActivity(activity);
   }

   emit updateThreadsStateSig(threads);
}

void Progress::Object::setItemActsThreadValue (Threads  thread,
                                               Steps    value)
{
#if defined(DEBUG_PROGRESS_DETAIL)
   printf("progress:%*s setItemActsThreadValue: [%d] (%llu:%llu)\n",
          indent, "", thread, value, threads[thread].tail);
#endif

   Steps  oldValue = threads[thread].value;

   threads[thread].value = value;
   emit updateThreadsStateSig(threads);

   Steps  newValue = itemValue + (value - oldValue);
   setItemValue(newValue);
}

void Progress::Object::endItemActThread (Threads  thread)
{
   threads[thread].state = Progress::Done;
   emit updateThreadsStateSig(threads);
}

void Progress::Object::retryItemActThread (Threads  thread)
{
   threads[thread].state = Progress::Retrying;

   // Revert any progress from the retrying thread.
   Steps newValue = itemValue - threads[thread].value;
   threads[thread].value = 0;
   setItemValue(newValue);

   emit updateThreadsStateSig(threads);
}

void Progress::Object::endItemActsThreads ()
{
#if defined(DEBUG_PROGRESS)
   printf("progress:       endItemActsThreads: {%llu}\n", itemTotalTail);
   indent = 6;
#endif

   threads = ThreadsState();
   emit updateThreadsStateSig(threads);
}

void Progress::Object::setItemValue (Steps  value)
{
   if (value > itemTail) {
      value = itemTail;
   }

   itemValue = value;

   Steps  itemLength      = itemTail;
   Steps  itemTotalLength = itemTotalTail - itemTotalStart;

   totalValue = (itemLength == 0
                 ? itemTotalStart + itemTotalLength
                 : itemTotalStart + value * itemTotalLength / itemLength);

   // See RATE_LIMIT_INTERVAL above for an explanation.
   if (lastSetValue.elapsed() >= RATE_LIMIT_INTERVAL
       || (value == itemTail && itemValue != itemTail)) {
      lastSetValue.start();

#if defined(DEBUG_PROGRESS_DETAIL)
      printf("progress:%*s setItemValue: {%llu/%llu}, (%llu/%llu)\n",
             indent, "",
             totalValue,
             totalTail,
             value,
             itemTail);
#endif

      emit setItemValueSig(totalValue, value);
   }

   updateGui();
}

void Progress::Object::setTotalValue (Steps  value)
{
   if (value > totalTail) {
      value = totalTail;
   }

   // See RATE_LIMIT_INTERVAL above for an explanation.
   if (lastSetValue.elapsed() < RATE_LIMIT_INTERVAL
       && (value != totalTail || totalValue == totalTail)) {
      return;
   }
   lastSetValue.start();

   totalValue = value;

#if defined(DEBUG_PROGRESS)
   printf("progress:%*s setTotalValue: {%llu/%llu}\n",
          indent, "", value, totalTail);
#endif

   emit setTotalValueSig(value);

   updateGui();
}

void Progress::Object::reset()
{
   totalTail = 1;

   itemTotalStart = 0;
   itemTotalTail  = 0;

   itemTail = 1;

   threads = ThreadsState();
   emit updateThreadsStateSig(threads);

#if defined(DEBUG_PROGRESS)
   printf("progress: reset\n");
#endif

   emit resetSig();

   updateGui();
}

void Progress::Object::updateGui()
{
   if (singleThread) {
      // Because there's no other thread updating the GUI, occasionally do
      // so manually here.
      if (lastGuiUpdate.elapsed() >= GUI_UPDATE_INTERVAL) {
         QEventLoop  eventLoop;
         eventLoop.processEvents(QEventLoop::ExcludeUserInputEvents);
         lastGuiUpdate.start();
      }
   }
}
